#pragma once

//Copyright (C) cocosimulator.org 2014
//generated by AmsterCHEM's COM CAPE-OPEN Wizard

// A COM smart pointer

template<class IFACE> class COMSmartPtr {

	IFACE *ptr;

	public:

	COMSmartPtr() {
		ptr=NULL;
	}

	COMSmartPtr(const COMSmartPtr<IFACE> &other) {
		ptr=other.ptr;
		if (ptr) ptr->AddRef();
	}

	COMSmartPtr(IFACE *object) {
		ptr=object;
		if (ptr) ptr->AddRef();
	}

	~COMSmartPtr() {
		if (ptr) ptr->Release();
	}

	const COMSmartPtr<IFACE> &operator=(const COMSmartPtr<IFACE> &object) {
		clear();
		if (object.ptr) {
			ptr=object.ptr;
			ptr->AddRef();
		}
		return *this;
	}

	const COMSmartPtr<IFACE> &operator=(IFACE*object) {
		clear();
		if (object) {
			ptr=object;
			ptr->AddRef();
		}
		return *this;
	}

	IFACE* operator->() {
  	  return ptr;
    }

	operator IFACE*() const 
	{
		return ptr;
	}

	IFACE& operator*() const
	{
		ASSERT(ptr,"Zero pointer dereferenced");
		return *ptr;
	}

	IFACE** operator&()
	{
		//we are likely being passed to QueryInterface, CoCreateInstance, etc
		ASSERT(!ptr,"Pointer is not zero while passed by reference");
		return &ptr;
	}

	void clear() {
		if (ptr) {
			ptr->Release();
			ptr=NULL;
		}
	}

	//attach and release do not alter the reference count

	void Attach(IFACE *object) {
		ptr=object;
	}

	IFACE * Detach() {
		IFACE *res=ptr;
		ptr=NULL;
		return res;
	}
};
