#include "WaterPackage.h"
#include "Water_DLL.h"
#include "COException.h"
#include "Water.h"
#include "PropertyMap.h"

//Copyright (C) cocosimulator.org 2014
//generated by AmsterCHEM's COM CAPE-OPEN Wizard


//WaterPackage.cpp - implementation of CAPE-OPEN WaterPackage class

WaterPackage::WaterPackage() : 
	CAPEOPENBASE(L"Water",L"IAPWS-97 based Property Package for calculation of water and steam",true),
	temperature(L"temperature"),
	pressure(L"pressure"),
	fraction(L"fraction"),
	phaseFraction(L"phaseFraction"),
	mole(L"mole"),
	mass(L"mass"),
	liquid(L"liquid"),
	vapor(L"vapor"),
	enthalpy(L"enthalpy"),
	entropy(L"entropy"),
	compoundsChecked(false) {
	//init compound list
	compIDs.MakeArray(1,VT_BSTR);
	compIDs.SetStringAt(0,CBStr(L"Water"));
	compFormulae.MakeArray(1,VT_BSTR);
	compFormulae.SetStringAt(0,CBStr(L"H2O"));
	compBoilTemps.MakeArray(1,VT_R8);
	CVariant var;
	propertyMap.GetConstValue(L"normalBoilingPoint",var);
	compBoilTemps.SetDoubleAt(0,var.Value().dblVal);
	compMolWts.MakeArray(1,VT_R8);
	propertyMap.GetConstValue(L"molecularWeight",var);
	compMolWts.SetDoubleAt(0,var.Value().dblVal);
	compCasnos.MakeArray(1,VT_BSTR);
	propertyMap.GetConstValue(L"casRegistryNumber",var);
	compCasnos.SetStringAt(0,var.Value().bstrVal);
	//init phase list
	phases.resize(2);
	phases[0]=L"Vapor";
	phases[1]=L"Liquid";
	phaseLabels.MakeArray(2,VT_BSTR);
	phaseLabels.SetStringAt(0,CBStr(phases[0]));
	phaseLabels.SetStringAt(1,CBStr(phases[1]));
	phaseAggState.MakeArray(2,VT_BSTR);
	phaseAggState.SetStringAt(0,CBStr(phases[0]));
	phaseAggState.SetStringAt(1,CBStr(phases[1]));
	phaseKeyComps.MakeArray(2,VT_BSTR);
	phaseKeyComps.SetStringAt(0,CBStr(L"Water"));
	phaseKeyComps.SetStringAt(1,CBStr(L"Water"));
	//init buffers
	scalar.MakeArray(1,VT_R8);
#ifdef _DEBUG
	//check consistency
	Water w;
	w.CheckConsistency();
#endif
}

STDMETHODIMP WaterPackage::GetCompoundConstant(/*[in]*/ VARIANT props,/*[in]*/ VARIANT compIds,/*[out,retval] */ VARIANT *compoundConstant) {
	if (!compoundConstant) return E_POINTER;
	BEGIN_TRY(L"ICapeThermoCompounds");
	int compCount;
	int i,j,index;
	wstring error;
	//check compound list
	CVariant compList(compIds,FALSE);
	if (!compList.CheckArray(VT_BSTR,error)) {
		throw COException(L"Invalid compound list: "+error);
	}
	if (compList.GetCount()==0) {
		//all compounds, i.e. water
		compCount=1;
	} else {
		for (i=0; i<compList.GetCount(); i++) {
			CBStr b=compList.GetStringAt(i);
			if (b.empty()) throw COException(L"Compound list containts empty string");
			if (!SameString(b,L"Water")) {
				error=L"Invalid compound in compound list: ";
				error+=b;
				throw COException(error);
			}
		}
		compCount=compList.GetCount();
	}
	//check property list
	CVariant propList(props,FALSE);
	if (!propList.CheckArray(VT_BSTR,error)) {
		throw COException(L"Invalid property list: "+error);
	}
	CVariant res;
	if (propList.GetCount()!=0) {
		//prep result array
		res.MakeArray(compCount*propList.GetCount(),VT_VARIANT);
		//loop over prop list
		index=0;
		for (i=0; i<propList.GetCount(); i++) {
			CVariant val;
			propertyMap.GetConstValue(propList.GetStringAt(i),val); //note that this will always return the value for compound water
			for (j=0; j<compCount; j++) {
				res.SetVariantAt(index,val.Value());
				index++;
			}
		}
	}
	*compoundConstant=res.ReturnValue();
	END_TRY(L"ICapeThermoCompounds");
}

STDMETHODIMP WaterPackage::GetCompoundList(/*[in,out]*/ VARIANT* compIds,/*[in,out]*/ VARIANT* formulae,/*[in,out]*/ VARIANT* names,/*[in,out]*/ VARIANT* boilTemps,/*[in,out]*/ VARIANT* molwts,/*[in,out]*/ VARIANT* casnos) {
	if ((!compIds)||(!formulae)||(!names)||(!boilTemps)||(!molwts)||(!casnos)) return E_POINTER;
	//all arguments are [in, out];
	VariantClear(compIds);
	VariantClear(formulae);
	VariantClear(names);
	VariantClear(boilTemps);
	VariantClear(molwts);
	VariantClear(casnos);
	//copy compound data (caller destroys data)
	*compIds=compIDs.Copy();
	*formulae=compFormulae.Copy();
	*names=compIDs.Copy(); //names and IDs are the same in this package
	*boilTemps=compBoilTemps.Copy();
	*molwts=compMolWts.Copy();
	*casnos=compCasnos.Copy();
	return NO_ERROR;
}

STDMETHODIMP WaterPackage::GetConstPropList(/*[out,retval] */ VARIANT *constPropList) {
	if (!constPropList) return E_POINTER;
	*constPropList=propertyMap.GetConstList().ReturnValue(); //ReturnValue is required for [out, retval], as caller destroys data
	return NO_ERROR;
}

STDMETHODIMP WaterPackage::GetNumCompounds(/*[out,retval] */ long int *numCompounds) {
	if (!numCompounds) return E_POINTER;
	*numCompounds=compIDs.GetCount();
	return NO_ERROR;
}

STDMETHODIMP WaterPackage::GetPDependentProperty(/*[in]*/ VARIANT props,/*[in]*/ double pressure,/*[in]*/ VARIANT compIds,/*[in,out]*/ VARIANT* propVals) {
	if (!propVals) return E_POINTER;
	BEGIN_TRY(L"ICapeThermoCompounds");
	wstring error;
	CVariant propList(props,FALSE);
	if (!propList.CheckArray(VT_BSTR,error)) throw COException(L"Invalid property list: "+error);
	if (propList.GetCount()>0) {
		throw COException(L"this package does not implement any pressure dependent properties",ECapeInvalidArgumentHR);
	}
	END_TRY(L"ICapeThermoCompounds");
}

STDMETHODIMP WaterPackage::GetPDependentPropList(/*[out,retval] */ VARIANT *pDependentPropList) {
	if (!pDependentPropList) return E_POINTER;
	pDependentPropList->vt=VT_EMPTY; //we do not have any 
	return NO_ERROR;
}

STDMETHODIMP WaterPackage::GetTDependentProperty(/*[in]*/ VARIANT props,/*[in]*/ double temperature,/*[in]*/ VARIANT compIds,/*[in,out]*/ VARIANT* propVals) {
	if (!propVals) return E_POINTER;
	BEGIN_TRY(L"ICapeThermoCompounds");
	int i,j,index,compCount;
	wstring error;
	Water w;
	double Psat,vVap,vLiq;
	//check valid T
	if (isMissing(temperature)) {
		throw COException(L"Missing temperature");
	}
	if (temperature<0) {
		throw COException(L"Zero or negative temperature");
	}
	//check compound list
	CVariant compList(compIds,FALSE);
	if (!compList.CheckArray(VT_BSTR,error)) {
		throw COException(L"Invalid compound list: "+error);
	}
	if (compList.GetCount()==0) {
		compCount=1; //all compounds in this package, i.e. water
	} else {
		for (i=0; i<compList.GetCount(); i++) {
			CBStr compName=compList.GetStringAt(i);
			if (compName.empty()) throw COException(L"Compound list containts empty string");
			if (!SameString(compName,L"Water")) {
				error=L"Invalid compound in compound list: ";
				error+=compName;
				throw COException(error);
			}
		}
		compCount=compList.GetCount();
	}
	//check property list
	CVariant propList(props,FALSE);
	if (!propList.CheckArray(VT_BSTR,error)) {
		throw COException(L"Invalid property list: "+error);
	}
	CVariant res;
	if (propList.GetCount()!=0) {
		//prep result array
		res.MakeArray(compCount*propList.GetCount(),VT_R8);
		//loop over prop list
		index=0;
		for (i=0; i<propList.GetCount(); i++) {
			double value;
			CBStr propName=propList.GetStringAt(i);
			switch (propertyMap.GetTDepProperty(propName)) {
			default:
				throw COException(L"Internal error: unimplemented temperature dependent property");
			case PropertyMap::SurfaceTensionSatLiquid:
				value=Water::surfaceTension(temperature);
				break;
			case PropertyMap::ThermalConductivityOfLiquid:
				Psat=Water::Psat(temperature);				
				w.SetStateTPX(temperature,Psat*1e-6,Water::TPLIQUID);
				value=w.thermalConductivity();
				break;
			case PropertyMap::ThermalConductivityOfVapor:
				Psat=Water::Psat(temperature);
				w.SetStateTPX(temperature,Psat*1e-6,Water::TPVAPOR);
				value=w.thermalConductivity();
				break;
			case PropertyMap::VaporPressure:
				value=Water::Psat(temperature)*1e6;
				break;
			case PropertyMap::VolumeChangeUponVaporization:
				Psat=Water::Psat(temperature);
				w.SetStateTPX(temperature,Psat*1e-6,Water::TPLIQUID);
				vLiq=w.volume();
				w.SetStateTPX(temperature,Psat*1e-6,Water::TPVAPOR);
				vVap=w.volume();
				value=(vVap-vLiq)*(Water::MOLWT*1e-3); //m3/mol
				break;
			case PropertyMap::VolumeOfLiquid:
				Psat=Water::Psat(temperature);
				w.SetStateTPX(temperature,Psat*1e-6,Water::TPLIQUID);
				value=w.volume()*(Water::MOLWT*1e-3);//m3/mol
				break;
			case PropertyMap::ViscosityOfLiquid:
				Psat=Water::Psat(temperature);
				w.SetStateTPX(temperature,Psat*1e-6,Water::TPLIQUID);
				value=w.viscosity();
				break;
			case PropertyMap::ViscosityOfVapor:
				Psat=Water::Psat(temperature);
				w.SetStateTPX(temperature,Psat*1e-6,Water::TPVAPOR);
				value=w.viscosity();
				break;
			case PropertyMap::IdealGasEnthalpy:
				//we return enthalpy at P=0
				//this property is just provided for reference state calculations for ideal gas reference state
				w.SetStateTPX(temperature,0,Water::TPVAPOR);
				value=w.enthalpy()*Water::MOLWT;//J/kg
				break;
			case PropertyMap::IdealGasEntropy: {
				//use external water ideal heat capacity correlation
				//this property is just provided for reference state calculations for ideal gas reference state
				const double A=36542.0;
				const double B=-34.8040008545;
				const double C=0.116820000112;
				const double D=-0.000130040003569;
				const double E=5.25470014168e-008;
				const double Pref=101325e-6;
				const double Tref=373.15; //vapor exists at Tref, Pref
				//RPP ideal gas heat capacity = A+B*T+C*T^2+D*T^3+E*T^4
				//ideal gas entropy: integral from Tref to T over Cp/T
				//parameters from Properties of Gases and Liquids 5th Edition, by Prausnitz
				value=temperature*(B+temperature*(0.5*C+temperature*(D/3.0+temperature*0.25*E)))+A*log(temperature)-
						(Tref*(B+Tref*(0.5*C+Tref*(D/3.0+Tref*0.25*E)))+A*log(Tref));
				value*=1e-3;
				//add the reference state value:
				w.SetStateTPX(Tref,Pref,Water::TPVAPOR);
				value+=w.entropy()*Water::MOLWT;//J/kg/K
				}
				break;
			}
			for (j=0; j<compCount; j++) {
				res.SetDoubleAt(index,value);
				index++;
			}
		}
	}
	*propVals=res.ReturnValue();
	END_TRY(L"ICapeThermoCompounds");
}

STDMETHODIMP WaterPackage::GetTDependentPropList(/*[out,retval] */ VARIANT *tDependentPropList) {
	if (!tDependentPropList) return E_POINTER;
	*tDependentPropList=propertyMap.GetTDepPropList().ReturnValue(); //ReturnValue is required for [out, retval], as caller destroys data
	return NO_ERROR;
}

STDMETHODIMP WaterPackage::GetNumPhases(/*[out,retval] */ long int *numPhases) {
	if (!numPhases) return E_POINTER;
	*numPhases=phaseLabels.GetCount(); 
	return E_NOTIMPL;
}

STDMETHODIMP WaterPackage::GetPhaseInfo(/*[in]*/ BSTR phaseLabel,/*[in]*/ BSTR phaseAttribute,/*[out,retval] */ VARIANT *value) {
	if (!value) return E_POINTER;
	BEGIN_TRY(L"ICapeThermoPhases");
	//check phase label
	int index=GetPhaseIndex(phaseLabel);
	//check phaseAttribute
	CBStr strVal;
	if (SameString(phaseAttribute,L"StateOfAggregation")) {
		strVal=phaseAggState.GetStringAt(index);
	} else if (SameString(phaseAttribute,L"KeyCompoundId")) {
		strVal=phaseKeyComps.GetStringAt(index);
	} else if ((SameString(phaseAttribute,L"ExcludedCompoundId"))||
		        (SameString(phaseAttribute,L"DensityDescription"))||
		        (SameString(phaseAttribute,L"UserDescription"))||
		        (SameString(phaseAttribute,L"TypeOfSolid"))) {
		//an ok attribute, but we do not have any value
	} else {
		//not valid
		wstring error;
		error=L"Invalid phase attribute: ";
		error+=(phaseAttribute)?phaseAttribute:L"NULL";
		throw COException(error,ECapeInvalidArgumentHR);
	}
	value->vt=VT_BSTR;
	value->bstrVal=strVal.Detach(); //will be destroyed by caller
	END_TRY(L"ICapeThermoPhases");
}

STDMETHODIMP WaterPackage::GetPhaseList(/*[in,out]*/ VARIANT* phaseLabels,/*[in,out]*/ VARIANT* stateOfAggregation,/*[in,out]*/ VARIANT* keyCompoundId) {
	if ((!phaseLabels)||(!stateOfAggregation)||(!keyCompoundId)) return E_POINTER;
	//all arguments are [in, out]
	VariantClear(phaseLabels);
	VariantClear(stateOfAggregation);
	VariantClear(keyCompoundId);
	//return copies of lists
	*phaseLabels=this->phaseLabels.Copy();
	*stateOfAggregation=phaseAggState.Copy();
	*keyCompoundId=phaseKeyComps.Copy();
	return NO_ERROR;
}

STDMETHODIMP WaterPackage::SetMaterial(/*[in]*/ LPDISPATCH material) {
	BEGIN_TRY(L"ICapeThermoCompounds");
	contextMat.clear();
	if (material) {
		if (FAILED(material->QueryInterface(IID_ICapeThermoMaterial,(LPVOID*)&contextMat))) {
			//no such interface
			throw COException(L"Object does not implement ICapeThermoMaterial");
		}
	}
	compoundsChecked=false; //must check at first use
	END_TRY(L"ICapeThermoCompounds");
}

STDMETHODIMP WaterPackage::UnsetMaterial() {
	contextMat.clear();
	compoundsChecked=false; 
	return NO_ERROR;
}

STDMETHODIMP WaterPackage::CalcAndGetLnPhi(/*[in]*/ BSTR phaseLabel,/*[in]*/ double temperature,/*[in]*/ double pressure,/*[in]*/ VARIANT moleNumbers,/*[in]*/ int fFlags,/*[in,out]*/ VARIANT* lnPhi,/*[in,out]*/ VARIANT* lnPhiDT,/*[in,out]*/ VARIANT* lnPhiDP,/*[in,out]*/ VARIANT* lnPhiDn) {
	BEGIN_TRY(L"ICapeThermoPropertyRoutine");
	throw COException(L"Fugacity coefficients are not implemented for this package",ECapeNoImplHR);
	END_TRY(L"ICapeThermoPropertyRoutine");
}

STDMETHODIMP WaterPackage::CalcSinglePhaseProp(/*[in]*/ VARIANT props,/*[in]*/ BSTR phaseLabel) {
	BEGIN_TRY(L"ICapeThermoPropertyRoutine");
	wstring err;
	double T,P;
	int i;
	//check context material
	if (contextMat==NULL) {
		throw COException(L"Context material object not set");
	}
	//check phase
	int phaseIndex=GetPhaseIndex(phaseLabel);
	//check prop list
	CVariant pList(props,FALSE);
	if (!pList.CheckArray(VT_BSTR,err)) {
		throw COException(L"Invalid property list: "+err);
	}
	//get temperature, pressure and composition
	GetTPX(phaseLabel,T,P);
	//alloc space
	names.resize(pList.GetCount());
	bases.resize(pList.GetCount());
	values.resize(pList.GetCount());
	//prep calculator
	Water w;
	w.SetStateTPX(T,P*1e-6,(phaseIndex==0)?Water::TPVAPOR:Water::TPLIQUID);
	//loop over all properties to calculate:
	for (i=0; i<pList.GetCount(); i++) {
		names[i]=pList.GetStringAt(i);
		switch (propertyMap.GetSinglePhaseProperty(names[i])) {
		case PropertyMap::Density:
			values[i]=1.0/(w.volume());//kg/m3
			bases[i]=mass;
			break;
		case PropertyMap::Enthalpy:
			values[i]=w.enthalpy()*1e3; //J/kg
			bases[i]=mass;
			break;
		case PropertyMap::Entropy:
			values[i]=w.entropy()*1e3; //J/kg/K
			bases[i]=mass;
			break;
		case PropertyMap::GibbsEnergy:
			values[i]=w.gibbs()*1e3; //J/kg
			bases[i]=mass;
			break;
		case PropertyMap::HeatCapacityCp:
			values[i]=w.cP()*1e3;//J/kg/K
			bases[i]=mass;
			break;
		case PropertyMap::HeatCapacityCv:
			values[i]=w.cV()*1e3;//J/kg/K
			bases[i]=mass;
			break;
		case PropertyMap::InternalEnergy:
			values[i]=w.energy()*1e3; //J/kg
			bases[i]=mass;
			break;
		case PropertyMap::MolecularWeight:
			values[i]=Water::MOLWT; //relative
			bases[i]=NULL;
			break;
		case PropertyMap::ThermalConductivity:
			values[i]=w.thermalConductivity();//W/m/K
			bases[i]=NULL;
			break;
		case PropertyMap::Volume:
			values[i]=w.volume();//m3/kg
			bases[i]=mass;
			break;
		case PropertyMap::Viscosity:
			values[i]=w.viscosity();//Pa.s
			bases[i]=NULL;
			break;
		default:
			throw COException(L"Internal error: unimplemented single phase property");
		}
	}
	//all properties have been calculated, set all properties on material object
	for (i=0; i<pList.GetCount(); i++) {
		SetSinglePhaseScalarProp(values[i],names[i],phaseLabel,bases[i]);
	}
	END_TRY(L"ICapeThermoPropertyRoutine");
}

STDMETHODIMP WaterPackage::CalcTwoPhaseProp(/*[in]*/ VARIANT props,/*[in]*/ VARIANT phaseLabels) {
	BEGIN_TRY(L"ICapeThermoPropertyRoutine");
	//check context material
	if (contextMat==NULL) {
		throw COException(L"Context material object not set");
	}
	//check phases
	int index1,index2;
	GetPhasePair(phaseLabels,index1,index2);
	//check prop list
	wstring error;
	CVariant pList(props,FALSE);
	if (!pList.CheckArray(VT_BSTR,error)) {
		throw COException(L"Invalid property list: "+error);
	}
	if (pList.GetCount()==0) {
		//done
		return NO_ERROR;
	}
	//get liquid temperature
	// (the only two phase property we support just depends on temperature)
	double T=GetSinglePhaseScalarProp(temperature,liquid,NULL);
	if (isMissing(T)) throw COException(L"Missing liquid temperature");
	//alloc space for prop names
	names.resize(pList.GetCount());
	values.resize(pList.GetCount());
	for (int i=0; i<pList.GetCount(); i++) {
		names[i]=pList.GetStringAt(i);
		switch (propertyMap.GetTwoPhaseProperty(names[i])) {
		case PropertyMap::SurfaceTension:
			values[i]=Water::surfaceTension(T);
			break;
		default:
			throw COException(L"Internal error: unimplemented two-phase property");
		}
	}
	//set results on material object
	for (int i=0; i<pList.GetCount(); i++) {
		scalar.SetDoubleAt(0,values[i]);
		HRESULT hr=contextMat->SetTwoPhaseProp(names[i],phaseLabels,NULL,scalar.Value());
		if (FAILED(hr)) {
			error=L"Failed to set value for ";
			error+=names[i];
			error+=L": ";
			error+=CO_Error(contextMat,hr);
			throw COException(error);
		}
	}
	END_TRY(L"ICapeThermoPropertyRoutine");
}

STDMETHODIMP WaterPackage::CheckSinglePhasePropSpec(/*[in]*/ BSTR property,/*[in]*/ BSTR phaseLabel,/*[out,retval] */ VARIANT_BOOL *supported) {
	//check support for the phase and property
	if (!supported) return E_POINTER;
	try {
		propertyMap.GetSinglePhaseProperty(property); //throws an exception if property does not exist; we do not care which property it is (we do not support liquid- or vapor-only properties)
		GetPhaseIndex(phaseLabel); //throws an exception in case of undefined phase, we do not care which phase it is
		*supported=VARIANT_TRUE;
	} catch (...) {
		*supported=VARIANT_FALSE;
	}
	return NO_ERROR;
}

STDMETHODIMP WaterPackage::CheckTwoPhasePropSpec(/*[in]*/ BSTR property,/*[in]*/ VARIANT phaseLabels,/*[out,retval] */ VARIANT_BOOL *supported) {
	//check support for the phase pair and property
	if (!supported) return E_POINTER;
	try {
		propertyMap.GetTwoPhaseProperty(property); //throws an exception if property does not exist; we do not care which property it is (we do not support liquid- or vapor-only properties)
		int index1,index2;
		GetPhasePair(phaseLabels,index1,index2); //throws an exception in case of undefined phase, we do not care which phase it is
		*supported=VARIANT_TRUE;
	} catch (...) {
		*supported=VARIANT_FALSE;
	}
	return NO_ERROR;
}

STDMETHODIMP WaterPackage::GetSinglePhasePropList(/*[out,retval] */ VARIANT *singlePhasePropList) {
	if (!singlePhasePropList) return E_POINTER;
	*singlePhasePropList=propertyMap.GetSinglePhasePropList().ReturnValue(); //ReturnValue is required for [out, retval], as caller destroys data
	return NO_ERROR;
}

STDMETHODIMP WaterPackage::GetTwoPhasePropList(/*[out,retval] */ VARIANT *twoPhasePropList) {
	if (!twoPhasePropList) return E_POINTER;
	*twoPhasePropList=propertyMap.GetTwoPhasePropList().ReturnValue(); //ReturnValue is required for [out, retval], as caller destroys data
	return NO_ERROR;
}

STDMETHODIMP WaterPackage::GetUniversalConstant(/*[in]*/ BSTR constantId,/*[out,retval] */ VARIANT *constantValue) {
	if (!constantValue) return E_POINTER;		
	constantValue->vt=VT_R8;
	constantValue->dblVal=propertyMap.GetUniversalConstant(constantId);
	return NOERROR;
}

STDMETHODIMP WaterPackage::GetUniversalConstantList(/*[out,retval] */ VARIANT *universalConstantList) {
	if (!universalConstantList) return E_POINTER;
	*universalConstantList=propertyMap.GetUniversalConstantList().ReturnValue(); //ReturnValue is required for [out, retval], as caller destroys data
	return NO_ERROR;
}

STDMETHODIMP WaterPackage::CalcEquilibrium(/*[in]*/ VARIANT specification1,/*[in]*/ VARIANT specification2,/*[in]*/ BSTR solutionType) {
	BEGIN_TRY(L"ICapeThermoEquilibriumRoutine");
	EqType type;
	BOOL vfVapor;
	BSTR VFBasis;
	bool haveT,haveP,haveVF;
	double T,P,X,H,S;
	//get equilibrium calculation spec
	GetEquilibriumType(specification1,specification2,solutionType,type,VFBasis,vfVapor);
	//get inputs, dependent on flash type
	haveT=haveP=haveVF=false;
	if (type==TP) {
		GetOverallTPX(T,P);
		haveT=haveP=true;
	} else {
		CheckOverallX();
		switch (type) {
		case TP:
		case TVF:
		case TH:
		case TS:
			//get overall temperature
			T=GetOverallScalarProp(temperature,NULL);
			//check valid T
			if (isMissing(T)) {
				throw COException(L"Temperature not specified or not finite");
			}
			if (T<0) {
				throw COException(L"Zero or negative temperature");
			}
			//T ok
			haveT=true;
		}
		switch (type) {
		case TP:
		case PVF:
		case PH:
		case PS:
			//get overall pressure
			P=GetOverallScalarProp(pressure,NULL);
			//check valid P
			if (isMissing(P)) {
				throw COException(L"Pressure not specified or not finite");
			}
			if (P<0) {
				throw COException(L"Zero or negative pressure");
			}
			//P ok
			haveP=true;
		}
		switch (type) {
		case TVF:
		case PVF:
		case HVF:
		case SVF:
			if (vfVapor) {
				//get vapor fraction
				X=GetSinglePhaseScalarProp(phaseFraction,vapor,VFBasis);
				if (isMissing(X)) {
					throw COException(L"Missing value for vapor fraction");
				}
			} else {
				//get liquid fraction
				X=GetSinglePhaseScalarProp(phaseFraction,liquid,VFBasis);
				if (isMissing(X)) {
					throw COException(L"Missing value for liquid fraction");
				}
				X=1.0-X; //make vapor fraction
			}
			if ((X<-.0000001)||(X>1.0000001)) {
				//out of range
				throw COException(L"Phase fraction out of [0,1] range");
			}
			//VF ok
			haveVF=true;
		}
		switch (type) {
		case TH:
		case PH:
		case HVF:
			//get overall enthalpy
			H=GetOverallScalarProp(enthalpy,mass);
			//check valid H
			if (isMissing(H)) {
				throw COException(L"Enthalpy not specified or not finite");
			}
		}
		switch (type) {
		case TS:
		case PS:
		case SVF:
			//get overall entropy
			S=GetOverallScalarProp(entropy,mass);
			//check valid S
			if (isMissing(S)) {
				throw COException(L"Entropy not specified or not finite");
			}
		}
	}
	//calculate the flash, at the end we need valid T, P, X(=vapor fraction)
	try {
		Water w;
		switch (type) {
		case TP: 
			//pressure, temperature
			w.SetStateTPX(T,P*1e-6,Water::TPAUTO);
			X=w.GetQuality();
			break;
		case TVF: //temperature, vapor fraction
			P=Water::Psat(T)*1e6;
			break;
		case PVF:	//pressure, vapor fraction
			T=Water::Tsat(P*1e-6);
			break;
		case PH: 
			//pressure, enthalpy
			w.SetStatePH(P*1e-6,H*1e-3);
			T=w.GetTemperature();
			X=w.GetQuality();
			break;
		case TH:
			//temperature, enthalpy
			w.SetStateTH(T,H*1e-3);
			P=w.GetPressure()*1e6;
			X=w.GetQuality();
			break;
		case PS: 
			//pressure, entropy
			w.SetStatePS(P*1e-6,S*1e-3);
			T=w.GetTemperature();
			X=w.GetQuality();
			break;
		case TS: 
			//temperature, entropy
			w.SetStateTS(T,S*1e-3);
			P=w.GetPressure()*1e6;
			X=w.GetQuality();
			break;
		case HVF: 
			//enthalpy, vapor fraction
			w.SetStateHVF(H*1e-3,X);
			P=w.GetPressure()*1e6;
			T=w.GetTemperature();
			break;
		case SVF: 
			//entropy, vapor fraction
			w.SetStateSVF(S*1e-3,X);
			P=w.GetPressure()*1e6;
			T=w.GetTemperature();
			break;
		}
	} catch (COException ex) {
		ex.Prepend(L"Flash calculation error: ");
		throw ex;
	}
	//flash calculation ok, set result on context material
	BOOL haveVap,haveLiq;
	int phaseCount,index;
	haveVap=((X>0)||(haveVF));
	haveLiq=((X<1.0)||(haveVF));
	//set present phases
	phaseCount=0;
	if (haveVap) {
		phaseCount++;
	}
	if (haveLiq) {
		phaseCount++;
	}
	CVariant phaseList,statList;
	phaseList.MakeArray(phaseCount,VT_BSTR);
	index=0;
	if (haveVap) {
		phaseList.SetStringAt(index++,vapor);
	}
	if (haveLiq) {
		phaseList.SetStringAt(index++,liquid);
	}
	statList.MakeArray(phaseCount,VT_I4);
	for (index=0; index<phaseCount; index++) {
		statList.SetLongAt(index,CAPE_ATEQUILIBRIUM);
	}
	HRESULT hr=contextMat->SetPresentPhases(phaseList.Value(),statList.Value());
	if (FAILED(hr)) {
		throw COException(L"Failed to set present phases: "+CO_Error(contextMat,hr));
	}
	if (X<0) {
		X=0;
	} else if (X>1.0) {
		X=1.0;
	}
	if (haveVap) {
		//set vapor VF, T, P, X
		SetSinglePhaseScalarProp(X,phaseFraction,vapor,mole);
		SetSinglePhaseScalarProp(T,temperature,vapor,NULL);
		SetSinglePhaseScalarProp(P,pressure,vapor,NULL);
		SetSinglePhaseScalarProp(1.0,fraction,vapor,mole);
	}
	if (haveLiq) {
		//set liquid VF, T, P, X
		SetSinglePhaseScalarProp(1.0-X,phaseFraction,liquid,mole);
		SetSinglePhaseScalarProp(T,temperature,liquid,NULL);
		SetSinglePhaseScalarProp(P,pressure,liquid,NULL);
		SetSinglePhaseScalarProp(1.0,fraction,liquid,mole);
	}
	if (!haveT) {
		//set overall T as it was not part of the flash spec
		SetOverallScalarProp(T,temperature,NULL);
	}
	if (!haveP) {
		//set overall P as it was not part of the flash spec
		SetOverallScalarProp(P,pressure,NULL);
	}
	END_TRY(L"ICapeThermoEquilibriumRoutine");
}

STDMETHODIMP WaterPackage::CheckEquilibriumSpec(/*[in]*/ VARIANT specification1,/*[in]*/ VARIANT specification2,/*[in]*/ BSTR solutionType,/*[out,retval] */ VARIANT_BOOL *checkEquilibriumSpecResult) {
	if (!checkEquilibriumSpecResult) return E_POINTER;
	try {
		EqType type;
		BOOL vfVapor;
		BSTR VFBasis;
		GetEquilibriumType(specification1,specification2,solutionType,type,VFBasis,vfVapor);
		*checkEquilibriumSpecResult=VARIANT_TRUE;
	} catch (...) {
		*checkEquilibriumSpecResult=VARIANT_FALSE;
	}
	return NO_ERROR;
}

STDMETHODIMP WaterPackage::get_parameters(/*[out,retval] */ LPDISPATCH *parameters) {
	*parameters=NULL;
	SetError(L"This object does not expose parameters",L"ICapeUtilities",L"get_parameters");
	return ECapeUnknownHR;
}

STDMETHODIMP WaterPackage::put_simulationContext(/*[in]*/ LPDISPATCH simulationContext) {
	//we are not using the simulation content.
	// typically one could use its ICapeDiagnostic, or ICapeCOSEUtilities; we don't
	return NO_ERROR;
}

STDMETHODIMP WaterPackage::Initialize() {
	//we do not depend on initialize
	return NO_ERROR;
}

STDMETHODIMP WaterPackage::Terminate() {
	//drop external references
	contextMat.clear();
	return NO_ERROR;
}

STDMETHODIMP WaterPackage::Edit() {
	//not supported
	SetError(L"Edit is not implemented by this object",L"ICapeUtilities",L"Edit");
	return ECapeNoImplHR;
}

/*CODE_WaterPackage_INSERT_LOCATION*/
